import os
import re
import logging
import shutil
import copy
import git
import hashlib
import datetime
import tarfile
import asyncio
import tempfile
from packaging.version import parse, Version
from enum import Enum

from . import (Spec, ENTRY_HEXSHA)
from pyrpm.spec import replace_macros


class ArchiveAlreadyThere(Exception):
    pass


class GitSourceError(Exception):
    pass


class RefType(Enum):
    SHA = 1
    BRANCH = 2
    TAG = 3


class GitSource:
    def __init__(self, workdir, log=None):
        self._pattern_version_tag = r"v{0,1}(\d+(\.\d+){0,2})"
        self.REGEX_VERSION_TAG = r"{}".format(self._pattern_version_tag)
        self.REGEX_DESCRIBE = r"{}(-(\d+)-g(\w+))*".format(
            self._pattern_version_tag)
        self.REGEX_SEMVER_LIKE = r"^\d+(\.\d+){0,2}$"
        self.REGEX_VERSION_SPECFILE = r"\d+(\.\d+){2}"
        self.REGEX_GITREV = r"g([a-f0-9]{7})"
        self.REGEX_GITREV_CHECKOUT = r"[a-f0-9]{7,40}"
        self.REGEX_ARCHIVE_EXTENSION = r"^(.*)-(\d+\.\d+.\d+(\+\d{8}(\+\d{6})?\+\d+\+g\w{7})*)\.(.+)$"

        self._workdir = workdir
        self._env_git = {"GIT_TERMINAL_PROMPT": "0"}
        if log:
            self.log = log
        else:
            self.log = logging.getLogger("GitSource")

        self._gitsourcedir = os.path.join(
            os.path.join(self._workdir, "gitsource"))
        self._copytags = None

        self._spec = None
        self._path_source = None
        self._ref_type = RefType.BRANCH
        self._archive_pattern = "%{name}-%{version}."
        self._archive_extension = None
        self._archive_version = None

    @property
    def current_commit(self):
        return self.repo_gitsource.head.commit

    #############
    # REGEX
    #############
    def _is_regex_gitrev_checkout(self, commit):
        return True if re.fullmatch(
            self.REGEX_GITREV_CHECKOUT, commit) else False

    def _regex_describe(self, describe):
        tag_re = re.fullmatch(self.REGEX_DESCRIBE, describe)
        if not tag_re:
            return (None, None, None)

        tagv = self.adjust_tagv(tag_re.group(1))

        # is commit tagged ?
        diff = tag_re.group(3)
        if not diff:
            return (tagv, None, None)

        nb_commit = tag_re.group(4)
        gcommit = tag_re.group(5)
        return (tagv, nb_commit, gcommit)

    def _regex_gitrev_specfile(self, version):
        name_re = re.search(self.REGEX_GITREV, version)
        if not name_re:
            return None

        return name_re.group(1)

    def _is_regex_version_specfile(self, version):
        return True if re.fullmatch(
            self.REGEX_VERSION_SPECFILE, version) else False

    # get commit checkout:
    #       load spec object if exists
    #       if no commit try to get commit from specfile
    def _get_commit_checkout(self, spec_content, commit):
        commit_checkout = commit if commit else None

        if not spec_content:
            self.log.debug("No spec file found!")
            self._spec = None
        else:
            self._spec = Spec.from_string(spec_content)

            # commit paramter override commit in specfile!!!
            if not commit_checkout:
                # if not mathching regex version: try to find the gcommit
                if not self._is_regex_version_specfile(self._spec.version):
                    commit_checkout = self._regex_gitrev_specfile(
                        self._spec.version)
                    if not commit_checkout:
                        self.log.warning(
                            "Cannot find matching version in specfile: reset commit checkout to master")

        commit_checkout = commit_checkout if commit_checkout else "master"
        self.log.info(
            "Will try to checkout to %s (commit_checkout)", commit_checkout)
        return commit_checkout

    # set path source:
    #       if no path source get the one from specfile
    def _set_path_source(self, path_source):
        if not path_source:
            if not self._spec or not self._spec.url:
                raise GitSourceError("Sources cannot be found!! \n"
                                     "Indeed, there is no url set in your "
                                     "specfile(URL: ) and "
                                     "no path source is passed in argument")
            self._path_source = self._spec.url
        else:
            self._path_source = path_source

    # shallow clone: depth=1, no-single-branch
    def _shallow_clone(self, commit_checkout):
        try:
            self.log.info("Trying shallow clone source from %s:%s into %s",
                          self._path_source, commit_checkout,
                          self._gitsourcedir)
            self.repo_gitsource = git.Repo.clone_from(
                self._path_source, self._gitsourcedir, branch=commit_checkout,
                env=self._env_git,
                depth=1, no_single_branch=True)
            self._is_shallow_cloned = True
            self.log.info("Successfull shallow cloned.")
        except git.exc.GitCommandError as e:
            os.rmdir(self._gitsourcedir)
            if e.args[1] == 128:
                raise GitSourceError(
                    "Invalid git reference: {}".format(commit_checkout))
            else:
                raise GitSourceError(
                    "Failed to clone {}. Not a valid git repository.".format(self._path_source))

    # shallow clone sha: commit_checkout is a sha
    #       shallow clone to master
    #       go_checkout will try to find deeper and deeper the matching sha
    #       at a point it will unshallow the repo
    def _shallow_clone_sha(self, commit_checkout):
        self._shallow_clone("master")
        if not self._go_checkout(commit_checkout):
            raise GitSourceError(
                "Invalid git reference: {}".format(commit_checkout))

    # fetch: update the repo + checkout to commit_checkout
    def _fetch(self, commit_checkout):
        try:
            self.log.info("Fetching repo from %s", self._path_source)
            self.repo_gitsource = git.Repo(self._gitsourcedir)

            if os.path.exists(self._gitsourcedir + "/.git/shallow"):
                self._is_shallow_cloned = True

            url = None
            self._get_remote_origin().fetch()

            self._set_ref_type(commit_checkout)

            if not self._go_checkout(commit_checkout):
                raise GitSourceError(
                    "Error in git command checkout %s" % commit_checkout)

            try:
                self._get_remote_origin().pull(rebase=True)
            except Exception:
                pass

        except (git.exc.GitCommandError, git.exc.InvalidGitRepositoryError):
            raise GitSourceError(
                "%s is not a valid git repository" % self._gitsourcedir)

    # set gitsource localdir: if path source is localdir: set gitsourcedir to it
    #       forbidden to clean in this case
    def _set_gitsource_localdir(self, clean):
        self._gitsourcedir = self._path_source
        try:
            self.repo_gitsource = git.Repo(self._gitsourcedir)
        except (git.exc.GitCommandError, git.exc.InvalidGitRepositoryError):
            raise GitSourceError(
                "%s is not a valid git repository" % self._gitsourcedir)

        if clean:
            raise GitSourceError(
                "Clean forbidden when --path_source is a local directory, "
                "aborting...")

    # set gitsource: try to set repo from gitsourcedir
    #       if failed: clean it
    #       check also that the remote is still valid: else clean
    def _set_gitsource(self, clean):
        try:
            self.repo_gitsource = git.Repo(self._gitsourcedir)
        except (git.exc.GitCommandError, git.exc.InvalidGitRepositoryError):
            clean = True

        # check remote is present in repo
        if not clean:  # if clean, gitsource will be removed anyway
            clean = True

            # check url from remote origin
            for url in self._get_remote_origin().urls:
                if self._path_source.strip() == url.strip():
                    clean = False

            # if still clean: warning
            if clean:
                self.log.warning(
                    "Remote: --path_source={} not found in remote in repo {},"
                    "gitsource will be cleaned".format(self._path_source, self._gitsourcedir))
                self.repo_gitsource = None
        return clean

    # checkout success: return True or False
    def checkout_success(self, checkout):
        try:
            self.repo_gitsource.git.checkout(checkout)
            self.log.info("Successful checkout to %s", checkout)
            return True
        except git.exc.GitCommandError:
            return False

    def _get_depth(self):
        return self.repo_gitsource.commit(
            rev="origin/master").count(first_parent=True)

    def _count_all(self):
        return self.repo_gitsource.head.commit.count(all=True)

    # git fetch depth: fetch increasing depth (shallow cloned)
    def _git_fetch_depth(self):
        depth = self._get_depth()
        depth += 10
        self.log.info("Fetching with depth %d", depth)
        self._get_remote_origin().fetch(depth=depth)
        return depth

    # git fetch depth callback: call callback after each fetch
    def _git_fetch_depth_callback(self, maxdepth, callback, *args):
        # first trying to fetch deeper and deeper: avoiding removing shallow
        # clone
        count = self._count_all()
        depth = self._get_depth()

        while maxdepth == 0 or depth < maxdepth:
            depth = self._git_fetch_depth()

            # if count is not updated
            new_count = self._count_all()
            if new_count <= count:
                break
            count = new_count

            if callback(*args):
                return True
        return False

    # go checkout: special checkout for shallow clone
    #       go deeper and deeper
    def _go_checkout(self, commit_checkout):
        if self.checkout_success(commit_checkout):
            return True

        if not self._is_shallow_cloned:
            return False

        if self._type_ref is not RefType.SHA:
            return False

        # first trying to fetch deeper and deeper: avoiding removing shallow
        # clone
        if self._git_fetch_depth_callback(
                100, self.checkout_success, commit_checkout):
            return True

        # final try: unshallow
        try:
            self._get_remote_origin().fetch("--unshallow")
            self.log.info(
                "Unshallowing repo because ref %s not found", commit_checkout)
        except git.GitCommandError:
            pass

        self._is_shallow_cloned = False

        return self.checkout_success(commit_checkout)

    # get remote origin: only work with origin remote
    def _get_remote_origin(self):
        return self.repo_gitsource.remotes.origin

    # check if commit checkout is available in ls remote
    def _git_check_ls_remote(self, options, commit_checkout):
        return True if git.cmd.Git().ls_remote(options, self._path_source,
                                               commit_checkout,
                                               env=self._env_git) else False

    # check if commit checkout is available in ls remote heads
    def _git_check_ls_remote_heads(self, commit_checkout):
        return self._git_check_ls_remote("--heads", commit_checkout)

    # check if commit checkout is available in ls remote tags
    def _git_check_ls_remote_tags(self, commit_checkout):
        return self._git_check_ls_remote("--tags", commit_checkout)

    # is branch: check if commit_checkout is a branch
    #       2 modes: from ls remote or from local repo
    def _is_branch(self, commit_checkout):
        if self.repo_gitsource:
            if commit_checkout not in self._get_remote_origin().refs:
                return False
        else:
            if not self._git_check_ls_remote_heads(commit_checkout):
                return False

        self._type_ref = RefType.BRANCH
        return True

    # is tag: check if commit_checkout is a tag
    #       2 modes: from ls remote or from local repo
    def _is_tag(self, commit_checkout):
        if self.repo_gitsource:
            if commit_checkout not in [str(tag)
                                       for tag in self.repo_gitsource.tags]:
                return False
        else:
            if not self._git_check_ls_remote_tags(commit_checkout):
                return False

        self._type_ref = RefType.TAG
        return True

    # set ref type: set the type of commit checkout
    def _set_ref_type(self, commit_checkout):
        # is branch ?
        is_set = self._is_branch(commit_checkout)

        # is tag ?
        if not is_set:
            is_set = self._is_tag(commit_checkout)

        # is looks like a sha ?
        if not is_set:
            is_set = self._is_regex_gitrev_checkout(commit_checkout)
            self._type_ref = RefType.SHA

        if not is_set:
            raise GitSourceError(
                "Invalid git reference {} for {}".format(
                    commit_checkout, self._path_source))

        self.log.info(
            "Commit checkout %s is set as a %s ref", commit_checkout,
            self._type_ref)

    def _need_redo_archive(self, force, old_version, old_extension):
        if force:
            return

        if self._spec.sha == self.repo_gitsource.head.commit.hexsha and self._archive_extension == old_extension:
            if self._archive_version and self._archive_version != old_version:
                return
            raise ArchiveAlreadyThere(
                "Commit in specfile {} and current commit after gitsource {} are the same, no need to pkg archive".format(
                    self._spec.sha, self.repo_gitsource.head.commit.hexsha))

    def _set_specfile(self, spec_content, previous_archive_name, force):
        specfile = Spec.from_string(spec_content)

        self._check_archive_specfile(specfile)

        self._spec.sha = specfile.sha

        if not previous_archive_name:
            return specfile

        previous_ext_re = re.search(
            self.REGEX_ARCHIVE_EXTENSION, previous_archive_name)

        if not previous_ext_re:
            return specfile

        # check old_name :! specfile name
        if previous_ext_re.group(1) != specfile.name:
            return specfile
        old_version = previous_ext_re.group(2)
        old_extension = previous_ext_re.group(5)
        self._need_redo_archive(force, old_version, old_extension)

        return specfile

    def _check_archive_specfile(self, specfile):
        # Ensure the archive is present in specfile
        for source in specfile.sources:
            if self._archive_pattern in source:
                self._archive_extension = source.replace(
                    self._archive_pattern, "")
                break
        if not self._archive_extension:
            raise GitSourceError("No archive find in specfile with the pattern {}".format(
                self._archive_pattern
            ))

    def adjust_tagv(self, tagv):
        tag_re = re.fullmatch(self.REGEX_VERSION_SPECFILE, tagv)
        if tag_re:
            return tagv

        # version is 1 digit: X
        tag_re = re.fullmatch(r"\d+", tagv)
        if tag_re:
            return "{}.0.0".format(tagv)

        # version is 2 digit: X
        tag_re = re.fullmatch(r"\d+\.\d+", tagv)
        if tag_re:
            return "{}.0".format(tagv)

        raise GitSourceError("tagv {} is not matching with regex: {}".format(
            tagv, self.REGEX_VERSION_SPECFILE))

    def _set_extended_version_tag(
            self, version, nb_commit, gcommit, detailed=False):
        commit_date = "{}{:02d}{:02d}".format(
            self.repo_gitsource.head.commit.committed_datetime.year,
            self.repo_gitsource.head.commit.committed_datetime.month,
            self.repo_gitsource.head.commit.committed_datetime.day)
        if detailed:
            commit_date = "{}+{:02d}{:02d}{:02d}".format(
                commit_date,
                self.repo_gitsource.head.commit.committed_datetime.hour,
                self.repo_gitsource.head.commit.committed_datetime.minute,
                self.repo_gitsource.head.commit.committed_datetime.second)

        self._version_tag = "{}+{}+{}+g{}".format(
            version, commit_date, nb_commit, gcommit)

    def _set_version_tag_from_describe(self):
        describe = None
        if not hasattr(self, "_args_describe_exclude"):
            self._args_describe_exclude = ["--tags"]

        try:
            describe = self.repo_gitsource.git.describe(
                self._args_describe_exclude)
        except git.exc.GitCommandError:
            # no decription found
            return False

        # get info from describe git matching our own criteria
        (self._version_tag, nb_commit, gcommit) = self._regex_describe(
            describe)

        # if no info found try again
        if not self._version_tag:
            REGEX_GET_TAG_FROM_DESCRIBE = r"-\d+-g\w+$"
            tag_to_exclude = re.sub(REGEX_GET_TAG_FROM_DESCRIBE, "", describe)
            self._args_describe_exclude += ["--exclude", tag_to_exclude]
            return self._set_version_tag_from_describe()

        # if info found but commit is not tagged
        if nb_commit and gcommit:
            self._set_extended_version_tag(
                self._version_tag, nb_commit, gcommit)
        self.log.info("Set version tag to %s", self._version_tag)
        return True

    def _looking_for_tag_with_git_describe(self, version):
        # looking for last commit tag
        self._version_tag = None
        if version:
            self._version_tag = version
            return

        if not self.repo_gitsource.tags:
            return

        # set version if it is found
        is_version_tag_set = self._set_version_tag_from_describe()
        # if version is set or not shallow cloned: return
        if is_version_tag_set or not self._is_shallow_cloned:
            return

        self.log.info("Not found tag matching regex: fetch deeper!")
        # shallow cloned: maybe tree is not enough complete to find a tag:
        # fetching
        self._git_fetch_depth_callback(
            0, self._set_version_tag_from_describe)

    def _get_change_log(self):
        today_str = datetime.date.today().strftime("%a %b %d %Y")
        commit_msg = ""
        lines_commit_msg = self.repo_gitsource.head.commit.message.split(
            os.linesep)
        # remove last empty element
        if lines_commit_msg[-1] == "":
            lines_commit_msg.pop(-1)
        for cmsg in lines_commit_msg:
            commit_msg += "- \t%s%s" % (cmsg, os.linesep)
        return "%s%s%s%s%s%s%s%s%s" % (os.linesep,
                                       "* %s IoT.bzh(iotpkg) <redpesk.list@iot.bzh> %s" % (
                                           today_str, self._version_tag),
                                       os.linesep,
                                       "- Upgrade version from source commit sha: %s" % self.repo_gitsource.head.commit.hexsha,
                                       os.linesep,
                                       "- Commit message:",
                                       os.linesep,
                                       commit_msg,
                                       os.linesep)

    def _get_new_spec_file_content(self, spec_content, no_changelog):
        # CHANGE LOG archive: prefill it
        is_changelog = True
        changelog = self._get_change_log()

        new_spec_file_content = ""
        # read file line by line
        for line in spec_content.splitlines():
            # rewrite version
            if re.search(r"^%s" % ENTRY_HEXSHA, line):
                continue
            name_re = re.search("[Vv]ersion:", line)
            if name_re:
                line = "%s %s%sVersion: %s" % (
                    ENTRY_HEXSHA, self.repo_gitsource.head.commit.hexsha, os.linesep, self._version_tag)

            # write new change log
            name_re = re.search("^%changelog", line)
            if name_re and not no_changelog:
                is_changelog = False
                line += changelog

            new_spec_file_content += "{}{}".format(line, os.linesep)

        # add a changelog if there is none
        if is_changelog and not no_changelog:
            new_spec_file_content += "{}%%changelog{}{}{}".format(
                os.linesep, os.linesep, changelog, os.linesep)
        return new_spec_file_content

    def _check_source(self, new_spec_file_content, archive_name):
        # check Source
        is_archive_in_sources = False
        new_spec = Spec.from_string(new_spec_file_content)
        for source in new_spec.sources:
            if archive_name == replace_macros(source, new_spec):
                is_archive_in_sources = True
        if not is_archive_in_sources:
            raise GitSourceError(
                "{} is not present in expanded sources ".format(archive_name))

    def _get_extensions(self):
        accepted_extensions_format = ["zip", "tgz", "tar.gz"]
        first_extension = None
        second_extension = None
        if self._archive_extension in accepted_extensions_format:
            first_extension = self._archive_extension
            second_extension = None
        elif self._archive_extension.startswith("tar"):
            first_extension = "tar"
            second_extension_re = re.search(
                r"\.(.*)$", self._archive_extension)
            if second_extension_re:
                second_extension = second_extension_re.group(1)
        else:
            raise GitSourceError(
                "Archive extension {} is not supported".format(
                    self._archive_extension))

        return (first_extension, second_extension)

    def _make_archive(self):
        # archive
        self.log.info("Making archive ...")
        (first_extension, second_extension) = self._get_extensions()

        prefix = "{}-{}".format(self._spec.name, self._version_tag)
        prefix_archive = "%s/" % prefix
        archive_name = "{}.{}".format(prefix, first_extension)

        archive_path = os.path.join(self._workdir, archive_name)
        archive_path_tmp = "%s~" % archive_path

        with open(archive_path_tmp, 'wb') as fp:
            self.repo_gitsource.archive(
                fp, prefix=prefix_archive, format=first_extension)

        if not second_extension:
            return (archive_name, archive_path, archive_path_tmp)

        archive_path_tmp_first_ext = archive_path_tmp
        archive_name = "{}.{}".format(prefix, self._archive_extension)
        archive_path = os.path.join(self._workdir, archive_name)
        archive_path_tmp = "%s~" % archive_path

        # extract in temporary dir
        with tempfile.TemporaryDirectory() as tmp_dir:
            with tarfile.open(archive_path_tmp_first_ext) as tar:
                tar.extractall(tmp_dir)

            os.remove(archive_path_tmp_first_ext)

            with tarfile.open(archive_path_tmp, "w:{}".format(second_extension)) as tar:
                tar.add(os.path.join(tmp_dir, prefix), arcname=prefix)

        return (archive_name, archive_path, archive_path_tmp)

    def gitsource(self, spec_content=None, commit=None,
                  path_source=None, clean=False):

        self._tags = {}
        self._is_shallow_cloned = False
        commit_checkout = self._get_commit_checkout(spec_content, commit)
        self._set_path_source(path_source)

        self.repo_gitsource = None
        # if local repository
        if os.path.isdir(self._path_source):
            self._set_gitsource_localdir(clean)
        elif os.path.isdir(self._gitsourcedir):
            clean = self._set_gitsource(clean)

        # clean source
        if clean:
            if os.path.isdir(self._gitsourcedir):
                self.log.debug("Remove %s", self._gitsourcedir)
                shutil.rmtree(self._gitsourcedir)
                self.repo_gitsource = None

        os.makedirs(self._gitsourcedir, exist_ok=True)

        # gitsource does not exist, clone the repo into it
        if not self.repo_gitsource:
            self._set_ref_type(commit_checkout)
            if self._type_ref is not RefType.SHA:
                self._shallow_clone(commit_checkout)
            else:
                # trying shallow clone with depth for finding sha
                self._shallow_clone_sha(commit_checkout)
        # gitsource exists, update it (fetch)
        else:
            self._fetch(commit_checkout)

        return self.repo_gitsource.head.commit

    def archive(self, spec_content, version=None, path_source=None,
                commit=None, clean=False, no_changelog=True,
                previous_archive_name=None, force=False):

        if version and not re.search(self.REGEX_SEMVER_LIKE, version):
            raise GitSourceError(
                "specfile version {} is not in the right format: {}".format(
                    version, self.REGEX_SEMVER_LIKE))
        self._archive_version = version

        self.gitsource(spec_content, path_source=path_source,
                       commit=commit, clean=clean)
        self.log.debug("Current commit is %s", self.repo_gitsource.head.commit)

        self._set_specfile(spec_content, previous_archive_name, force)

        self._looking_for_tag_with_git_describe(version)

        # no tag found: version is 0.0.0
        if not self._version_tag:
            self._set_extended_version_tag(
                "0.0.0", "0", self.repo_gitsource.head.commit.hexsha[0:7], True)

        # all is indexed ?
        if self.repo_gitsource.index.diff(None):
            self._version_tag = "%s+dirty" % self._version_tag

        self.log.debug("Version is %s", self._version_tag)

        (archive_name, archive_path, archive_path_tmp) = self._make_archive()

        new_spec_file_content = self._get_new_spec_file_content(
            spec_content, no_changelog)

        self._check_source(new_spec_file_content, archive_name)

        # rename temporary archive
        if os.path.isfile(archive_path):
            os.remove(archive_path)
        os.rename(archive_path_tmp, archive_path)

        self.log.info(archive_name)
        return archive_path, new_spec_file_content, self.repo_gitsource.head.commit.hexsha
